import java.io.PrintStream;

import org.apache.bcel.Constants;
import org.apache.bcel.generic.BasicType;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.GETFIELD;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionFactory;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.Type;

/**
 * Given instruction list and getfield instruction position this class is
 * responsible for modifying the list so that before getfiled instruction a
 * message describing the instruction will be printed.
 * 
 * @author Kuba Madej
 * 
 */
public class InstructionBuilder {
	/**
	 * System's filed name containing PrintStream instance associated with
	 * stdout
	 */
	private static final String OUT_FIELD = "out";

	/**
	 * PrintStream's println method name
	 */
	private static final String PRINTLN_METHOD = "println";

	/**
	 * StringBuilder's toString method name
	 */
	private static final String TO_STRING_METHOD = "toString";

	/**
	 * StringBuilder's append method name
	 */
	private static final String APPEND_METHOD = "append";

	/**
	 * StringBuilder class name
	 */
	private static final String STRING_BUILDER_CLASS = StringBuilder.class
			.getName();

	/**
	 * New line string
	 */
	private static final String NEWLINE = "\n";

	/**
	 * Message prefix
	 */
	private static final String MESSAGE_PREFIX = "Before getfield:" + NEWLINE;

	/**
	 * Tabulator string
	 */
	private static final String TAB = "    ";

	/**
	 * Instruction list to insert instructions to
	 */
	private final InstructionList instructionList;

	/**
	 * Instruction handle pointing to the getfield instruction
	 */
	private final InstructionHandle getfieldInstructionHandle;

	/**
	 * ConstantPoolGen of the class that containts instruction list
	 */
	private final ConstantPoolGen constantPoolGen;

	/**
	 * getfield instruction
	 */
	private final GETFIELD getfieldInstruction;

	/**
	 * Instruction factory to be used
	 */
	private final InstructionFactory instructionFactory;

	/**
	 * Constructs new builder instance
	 * 
	 * @param instructionList
	 * @param instructionHandle
	 * @param instructionFactory
	 */
	public InstructionBuilder(final InstructionList instructionList,
			final InstructionHandle instructionHandle,
			final ConstantPoolGen constantPoolGen) {
		this.instructionList = instructionList;
		this.getfieldInstructionHandle = instructionHandle;
		this.constantPoolGen = constantPoolGen;
		getfieldInstruction = (GETFIELD) instructionHandle.getInstruction();
		instructionFactory = new InstructionFactory(constantPoolGen);
	}

	/**
	 * Creates necessary instructions and inserts them to instruction list
	 * passed in the constructor
	 * 
	 */
	public void build() {
		createStaticPrint();
		createDynamicPrint();
	}

	/**
	 * Creates and inserts instructions responsible for the static part of the
	 * message (containing data known before runtime)
	 * 
	 */
	private void createStaticPrint() {
		final StringBuilder builder = new StringBuilder(MESSAGE_PREFIX);
		builder.append(TAB);
		builder.append(getfieldInstruction.getReferenceType(constantPoolGen));
		builder.append(NEWLINE + TAB);
		builder.append(getfieldInstruction.getFieldType(constantPoolGen));
		builder.append(NEWLINE + TAB);
		builder.append(getfieldInstruction.getFieldName(constantPoolGen));

		instructionList.insert(getfieldInstructionHandle, instructionFactory
				.createPrintln(builder.toString()));
	}

	/**
	 * Creates and inserts instructions responsible for the dynamic part of the
	 * message (containing data known at runtime)
	 */
	private void createDynamicPrint() {
		duplicateReference();
		getSystemOut();
		swapStackTop();
		createNewStringBuilder();
		duplicateReference();
		callStringBuilderMethod(Constants.CONSTRUCTOR_NAME, BasicType.VOID,
				null, Constants.INVOKESPECIAL);
		getTabConstant();
		callStringBuilderAppendMethod(ObjectType.STRING);
		swapStackTop();
		getField();
		callStringBuilderAppendMethod(getfieldInstruction
				.getFieldType(constantPoolGen));
		callStringBuilderMethod(TO_STRING_METHOD, ObjectType.STRING, null,
				Constants.INVOKEVIRTUAL);
		callMethod(PrintStream.class.getName(), PRINTLN_METHOD, BasicType.VOID,
				ObjectType.STRING, Constants.INVOKEVIRTUAL);
	}

	/**
	 * Creates and inserts getfield instruction
	 * 
	 */
	private void getField() {
		insertInstruction(getfieldInstruction);
	}

	/**
	 * Creates and inserts ldc instruction
	 * 
	 */
	private void getTabConstant() {
		insertInstruction(instructionFactory.createConstant(TAB));
	}

	/**
	 * Creates and inserts 'invoke' instruction
	 * 
	 * @param className
	 *            class name containing method to be called
	 * @param methodName
	 *            name of the method to be called
	 * @param returnType
	 *            return type, void in case of a constructor
	 * @param argumentType
	 *            argument type or null if the method has no arguments
	 * @param invokeType
	 *            invoke instruction type
	 */
	private void callMethod(final String className, final String methodName,
			final Type returnType, final Type argumentType,
			final short invokeType) {
		insertInstruction(instructionFactory.createInvoke(className,
				methodName, returnType,
				argumentType != null ? new Type[] { argumentType }
						: new Type[0], invokeType));
	}

	/**
	 * Creates and inserts method call to StringBuilder class
	 * 
	 * @param methodName
	 *            method name
	 * @param returnType
	 *            return type
	 * @param argumentType
	 *            argument type or null if the method has no arguments
	 * @param invokeType
	 *            invoke instruction type
	 */
	private void callStringBuilderMethod(final String methodName,
			final Type returnType, final Type argumentType,
			final short invokeType) {
		callMethod(STRING_BUILDER_CLASS, methodName, returnType, argumentType,
				invokeType);
	}

	/**
	 * Creates and inserts StringBuilder method invocation
	 * 
	 * @param argumentType
	 *            argument type
	 */
	private void callStringBuilderAppendMethod(Type argumentType) {
		if (argumentType.equals(BasicType.BYTE)
				|| argumentType.equals(BasicType.SHORT))
			argumentType = BasicType.INT;
		callStringBuilderMethod(APPEND_METHOD, new ObjectType(
				STRING_BUILDER_CLASS), argumentType, Constants.INVOKEVIRTUAL);
	}

	/**
	 * Creates and inserts 'new' instruction that constructs new StringBuilder
	 * instance
	 * 
	 */
	private void createNewStringBuilder() {
		insertInstruction(instructionFactory.createNew(new ObjectType(
				STRING_BUILDER_CLASS)));
	}

	/**
	 * Creates and inserts 'swap' instruction
	 * 
	 */
	private void swapStackTop() {
		insertInstruction(InstructionFactory.SWAP);
	}

	/**
	 * Creates and inserts 'getstatic' instruction that fetches System.out
	 * reference
	 * 
	 */
	private void getSystemOut() {
		insertInstruction(instructionFactory.createGetStatic(System.class
				.getName(), OUT_FIELD, new ObjectType(PrintStream.class
				.getName())));
	}

	/**
	 * Creates and inserts 'dup' instruction
	 * 
	 */
	private void duplicateReference() {
		insertInstruction(InstructionFactory.DUP);
	}

	/**
	 * Inserts a single instruction into instruction list
	 * 
	 * @param instruction
	 *            instruction to be inserted
	 */
	private void insertInstruction(final Instruction instruction) {
		insertInstructionList(new InstructionList(instruction));
	}

	/**
	 * Inserts given instructions into instruction list
	 * 
	 * @param list
	 *            instructions to be inserted
	 */
	private void insertInstructionList(final InstructionList list) {
		instructionList.insert(getfieldInstructionHandle, list);
	}
}
