import java.io.IOException;
import java.util.Iterator;

import org.apache.bcel.Repository;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.util.InstructionFinder;

/**
 * This class performs transformation of the given class. Any getfield
 * istruction found is preceeded with printing information about instruction's
 * operand.
 * 
 * @author Kuba Madej
 * 
 */
public class Transform {
	/**
	 * Java class file extension
	 */
	private static final String CLASS_EXTENSION = ".class";

	/**
	 * Regex pattern matching any method invocation
	 */
	private static final String GETFIELD = "getfield";

	/**
	 * Return code of the preoces in case of failure
	 */
	private static final int FAILURE_CODE = 1;

	/**
	 * Class to be transformed
	 */
	private JavaClass javaClass;

	/**
	 * Constant pool generator of class to be transformed
	 */
	private final ConstantPoolGen constantPoolGen;

	/**
	 * Constructs new transformer
	 * 
	 * @param className
	 *            name of the class to be transformed
	 */
	public Transform(final String className) {
		try {
			javaClass = Repository.lookupClass(className);
		} catch (ClassNotFoundException e) {
			System.err.println("Class not found");
			System.exit(FAILURE_CODE);
		}

		constantPoolGen = new ConstantPoolGen(javaClass.getConstantPool());
	}

	/**
	 * Starts the transofmer
	 * 
	 * @param args
	 *            arguments array
	 */
	public static void main(final String[] args) {
		if (args.length < 1) {
			System.err.println("Usage: java Transform <classname>");
			System.exit(0);
		}
		new Transform(args[0]).transform();
	}

	/**
	 * Performs the actual transformation
	 * 
	 */
	public void transform() {
		final Method[] methods = javaClass.getMethods();

		for (int i = 0; i < methods.length; i++) {
			methods[i] = transform(new MethodGen(methods[i], javaClass
					.getClassName(), constantPoolGen));

		}

		javaClass.setConstantPool(constantPoolGen.getFinalConstantPool());

		try {
			javaClass.dump(javaClass.getClassName() + CLASS_EXTENSION);
		} catch (IOException e) {
			System.err.println("Couldn't write class file");
			System.exit(FAILURE_CODE);
		}
	}

	/**
	 * Transforms a single method
	 * 
	 * @param methodGen
	 *            MethodGen instance of the method to be tranformed
	 * @return transformed method
	 */
	private Method transform(final MethodGen methodGen) {
		final InstructionList instructionList = methodGen.getInstructionList();
		if (instructionList == null) {
			// native or abstract method
			return methodGen.getMethod();
		}

		final InstructionFinder instructionFinder = new InstructionFinder(
				instructionList);

		for (final Iterator iter = instructionFinder.search(GETFIELD); iter
				.hasNext();) {
			final InstructionHandle[] match = (InstructionHandle[]) iter.next();
			final InstructionHandle getfieldInstructionHandle = match[0];

			new InstructionBuilder(instructionList, getfieldInstructionHandle,
					constantPoolGen).build();
		}

		methodGen.setMaxStack();
		return methodGen.getMethod();
	}
}
