import java.io.IOException;
import java.util.Iterator;

import org.apache.bcel.Constants;
import org.apache.bcel.Repository;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.generic.ASTORE;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.INVOKEINTERFACE;
import org.apache.bcel.generic.INVOKESPECIAL;
import org.apache.bcel.generic.INVOKEVIRTUAL;
import org.apache.bcel.generic.ISTORE;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionFactory;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.generic.InvokeInstruction;
import org.apache.bcel.generic.LocalVariableGen;
import org.apache.bcel.generic.LocalVariableInstruction;
import org.apache.bcel.generic.MethodGen;
import org.apache.bcel.generic.NOP;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.ReturnInstruction;
import org.apache.bcel.generic.TargetLostException;
import org.apache.bcel.generic.Type;

/**
 * This class performs transformation of the given class. It inlines method
 * invocations performed on another given class.
 *
 * @author Kuba Madej
 *
 */
public class Transform {

	/**
	 * Constructor method name
	 */
	private static final String CONSTRUCTOR_METHOD_NAME = "<init>";

	/**
	 * Return code of the preoces in case of failure
	 */
	private static final int FAILURE_CODE = 1;

	/**
	 * Class file extension
	 */
	private static final String CLASS_FILE_EXTENSION = ".class";

	/**
	 * Class to be transformed
	 */
	private JavaClass classToTransform;

	/**
	 * Constant pool generator of class to be transformed
	 */
	private final ConstantPoolGen constantPoolGen;

	/**
	 * Class to be inlined
	 */
	private JavaClass classToInline;

	/**
	 * Constant pool generator of class to be inlined
	 */
	private ConstantPoolGen inlineConstantPoolGen;

	/**
	 * Constructs new transformer
	 *
	 * @param classNameToBeInlined
	 *            name of the class to be transformed
	 * @param classNameToInline
	 */
	public Transform(final String classNameToBeInlined,
			final String classNameToInline) {
		try {
			classToTransform = Repository.lookupClass(classNameToBeInlined);
			classToInline = Repository.lookupClass(classNameToInline);
		} catch (ClassNotFoundException e) {
			System.err.println("Class not found");
			System.exit(FAILURE_CODE);
		}

		constantPoolGen = new ConstantPoolGen(classToTransform
				.getConstantPool());
		inlineConstantPoolGen = new ConstantPoolGen(classToInline
				.getConstantPool());
	}

	/**
	 * Starts the transofmer
	 *
	 * @param args
	 *            arguments array
	 * @throws ClassNotFoundException
	 */
	public static void main(final String[] args) throws ClassNotFoundException {
		if (args.length < 2) {
			System.err.println("Usage: java Transform <classname> <classname>");
			System.exit(0);
		}
		new Transform(args[0], args[1]).transform();
	}

	/**
	 * Performs the actual transformation
	 *
	 * @throws ClassNotFoundException
	 *
	 */
	public void transform() throws ClassNotFoundException {
		final Method[] methods = classToTransform.getMethods();

		for (int i = 0; i < methods.length; i++) {
			final Method method = methods[i];
			if (!(method.isNative() || method.isAbstract())) {
				methods[i] = transform(new MethodGen(method, classToTransform
						.getClassName(), constantPoolGen));
			}

		}

		classToTransform
				.setConstantPool(constantPoolGen.getFinalConstantPool());

		try {
			classToTransform.dump(classToTransform.getClassName()
					+ CLASS_FILE_EXTENSION);
		} catch (IOException e) {
			System.err.println("Couldn't write class file");
			System.exit(FAILURE_CODE);
		}
	}

	/**
	 * Transforms a single method
	 *
	 * @param methodGen
	 *            MethodGen instance of the method to be tranformed
	 * @return transformed method
	 * @throws ClassNotFoundException
	 */
	private Method transform(final MethodGen methodGen)
			throws ClassNotFoundException {
		final InstructionList instructionList = methodGen.getInstructionList();

		for (Iterator iter = instructionList.iterator(); iter.hasNext();) {
			final InstructionHandle invokeInstructionHandle = (InstructionHandle) iter
					.next();
			final InstructionHandle nextInstructionHandle = invokeInstructionHandle
					.getNext();

			final Instruction instruction = invokeInstructionHandle
					.getInstruction();

			if (instruction instanceof InvokeInstruction
					&& !(instruction instanceof INVOKESPECIAL)) {
				InvokeInstruction invokeInstruction = (InvokeInstruction) instruction;

				if (invokeInstruction.getReferenceType(constantPoolGen).equals(
						new ObjectType(classToInline.getClassName()))) {

					final Method methodToInline = findMethodToInline(invokeInstruction);

					final MethodGen inlinedMethodGen = new MethodGen(
							methodToInline, classToInline.getClassName(),
							inlineConstantPoolGen);

					final int localsStart = fixLocalVariablesSlots(methodGen,
							inlinedMethodGen);

					final InstructionList instructionsToInsert = inlinedMethodGen
							.getInstructionList().copy();

					instructionsToInsert
							.replaceConstantPool(new ConstantPoolGen(
									classToInline.getConstantPool()),
									constantPoolGen);

					instructionList.append(invokeInstructionHandle,
							instructionsToInsert);

					stackToLocals(instructionList, invokeInstructionHandle,
							instruction, inlinedMethodGen, localsStart);

					for (InstructionHandle handle = invokeInstructionHandle; handle != nextInstructionHandle; handle = handle
							.getNext()) {
						final Instruction instructionToInsert = handle
								.getInstruction();

						if (instructionToInsert instanceof ReturnInstruction) {

							handle.setInstruction(new NOP());

							instructionList.append(handle, InstructionFactory
									.createBranchInstruction(Constants.GOTO,
											nextInstructionHandle));
						}

						if (instructionToInsert instanceof INVOKESPECIAL) {
							INVOKESPECIAL invokeSpecial = (INVOKESPECIAL) instructionToInsert;

							if (!invokeSpecial.getMethodName(constantPoolGen)
									.equals(CONSTRUCTOR_METHOD_NAME))
								handle.setInstruction(new INVOKEVIRTUAL(
										invokeSpecial.getIndex()));
						}
					}

					try {
						instructionList.delete(invokeInstructionHandle);
					} catch (TargetLostException e) {
						// safe to ignore
					}

				}
			}
		}

		methodGen.setMaxLocals();
		methodGen.setMaxStack();
		return methodGen.getMethod();
	}

	/**
	 * Generates code that converts stack contents to local variables.
	 *
	 * @param instructionList
	 *            instruction list to modify
	 * @param invokeInstructionHandle
	 *            invoke instruction handle
	 * @param instruction
	 *            instruction
	 * @param inlinedMethodGen
	 *            inlined method gen
	 * @param localsStart
	 *            lowest index of local variables used by code to be inlined
	 */
	private void stackToLocals(final InstructionList instructionList,
			final InstructionHandle invokeInstructionHandle,
			final Instruction instruction, final MethodGen inlinedMethodGen,
			final int localsStart) {
		final boolean instanceMethod = instruction instanceof INVOKEINTERFACE
				|| instruction instanceof INVOKEVIRTUAL;
		final int argumentsShift = instanceMethod ? 1 : 0;
		final Type[] argumentTypes = inlinedMethodGen.getArgumentTypes();
		if (instanceMethod) {
			instructionList.append(invokeInstructionHandle, new ASTORE(
					localsStart));
		}
		for (int i = localsStart + argumentsShift; i < localsStart
				+ argumentTypes.length + argumentsShift; i++) {
			final Type type = argumentTypes[i - localsStart - argumentsShift];

			if (type.equals(Type.INT)) {
				instructionList.append(invokeInstructionHandle, new ISTORE(i));
			} else {
				instructionList.append(invokeInstructionHandle, new ASTORE(i));
			}
		}
	}

	/**
	 * Returns method to be inlined.
	 *
	 * @param invokeInstruction
	 *            invoke instruction
	 * @return method to be inlined
	 */
	private Method findMethodToInline(InvokeInstruction invokeInstruction) {
		for (Method method : classToInline.getMethods()) {
			if (method.getSignature().equals(
					invokeInstruction.getSignature(constantPoolGen))
					&& method.getName().equals(
							invokeInstruction.getMethodName(constantPoolGen))) {

				return method;
			}
		}
		throw new RuntimeException(
				"Couldn't find method matching to method invocation that occured");
	}

	/**
	 * Changes indexes of local variables used by inlined code so that they do
	 * not collide with existing method code.
	 *
	 * @param methodGen
	 *            MethodGen of method being transformed
	 * @param inlinedMethodGen
	 *            MethodGen of method being inlined
	 * @return new start index of local variables used by inlined code
	 */
	private int fixLocalVariablesSlots(final MethodGen methodGen,
			final MethodGen inlinedMethodGen) {
		final LocalVariableGen[] localVariables = methodGen.getLocalVariables();
		final LocalVariableGen[] inlinedLocalVariables = inlinedMethodGen
				.getLocalVariables();

		int maxSlotNumber = 0;
		int minInlinedSlotNumber = Integer.MAX_VALUE;

		for (LocalVariableGen gen : localVariables) {
			maxSlotNumber = Math.max(maxSlotNumber, gen.getIndex());
		}

		for (LocalVariableGen gen : inlinedLocalVariables) {
			minInlinedSlotNumber = Math.min(minInlinedSlotNumber, gen
					.getIndex());
		}

		if (minInlinedSlotNumber > maxSlotNumber) {
			return minInlinedSlotNumber;
		}

		final int shift = maxSlotNumber - minInlinedSlotNumber + 1;

		final InstructionList instructionList = inlinedMethodGen
				.getInstructionList();

		for (Iterator iter = instructionList.iterator(); iter.hasNext();) {
			InstructionHandle handle = (InstructionHandle) iter.next();

			final Instruction instruction = handle.getInstruction();

			if (instruction instanceof LocalVariableInstruction) {
				LocalVariableInstruction localVariableInstruction = (LocalVariableInstruction) instruction;

				localVariableInstruction.setIndex(localVariableInstruction
						.getIndex()
						+ shift);
			}
		}

		return minInlinedSlotNumber + shift;
	}
}
